#!/usr/bin/perl

# Copyright (C) 2005 John J. Chew, III <jjchew@math.utoronto.ca>
# All Rights Reserved

package TSH::Command::RandomPair;

use strict;
use warnings;

use TSH::Player;
use TSH::Utility qw(Debug);

our (@ISA) = qw(TSH::PairingCommand);

=pod

=head1 NAME

TSH::Command::RandomPair - implement the C<tsh> RandomPair command

=head1 SYNOPSIS

  my $command = new TSH::Command::RandomPair;
  my $argsp = $command->ArgumentTypes();
  my $helptext = $command->Help();
  my (@names) = $command->Names();
  $command->Run($tournament, @parsed_arguments);
  
=head1 ABSTRACT

TSH::Command::RandomPair is a subclass of TSH::Command.

=cut

=head1 DESCRIPTION

=over 4

=cut

sub initialise ($$$$);
sub new ($);
sub Run ($$@);

=item $parserp->initialise()

Used internally to (re)initialise the object.

=cut

sub initialise ($$$$) {
  my $this = shift;
  my $path = shift;
  my $namesp = shift;
  my $argtypesp = shift;

  $this->{'help'} = <<'EOF';
Use the RandomPair command to add a round of random pairings.
Although it might seem that you might not need to specify a round
on which to base the pairings, that value is required in case we
need to decide who gets a bye.
EOF
  $this->{'names'} = [qw(rp randompair)];
  $this->{'argtypes'} = [qw(Repeats BasedOnRound Division)];
# print "names=@$namesp argtypes=@$argtypesp\n";

  return $this;
  }

sub new ($) { return TSH::Utility::new(@_); }

=item $command->Run($tournament, @parsed_args)

Should run the command in the context of the given
tournament with the specified parsed arguments.

=cut

sub Run ($$@) { 
  my $this = shift;
  my $tournament = shift;
  my ($repeats, $sr, $dp) = @_;
  my $sr0 = $sr - 1;

  $dp->CheckRoundHasResults($sr0) or return 0;
  $tournament->TellUser('irsok', $dp->Name(), $sr, $repeats);
  my $exagony = $config::exagony || ($config::initial_exagony && $dp->LastPairedRound0() == -1);
  if ($exagony) { $tournament->TellUser('irsem'); }
  my $psp = $dp->GetRegularUnpaired($sr0);
  unless (@$psp) { $tournament->TellUser('ealpaird'); return; }
  die "Assertion failed" unless @$psp % 2 == 0;

  if (TSH::Player::PairGRT($psp,
    # opponent preference ranking
    # $psp is arg 0
    # $pindex is arg 1
    # $oppindex is arg 2
    sub {
      my $p = $_[0][$_[1]];
      my $pid = $p->ID();
      my $o = $_[0][$_[2]];
      my $oid = $o->ID();
      my $lastoid = ($p->OpponentID(-1)||-1);
      my $thisrep = $p->Repeats($oid); 
      my $sameopp = ($oid == $lastoid);
      my $distance = $p->Random() - $o->Random();
      my $pairsvr = $config::track_firsts ? 2-abs(($p->{'p1'}-$p->{'p2'} <=> 0)  -($o->{'p1'}-$o->{'p2'} <=> 0)) : 0;

      Debug 'GRT', 'pref %d-%d rep=%d prev=%d svr=%d dist=%d', $pid, $oid, $thisrep, $sameopp, $pairsvr, $distance;
      pack('NCCNN',
	$distance, # prefer random opponents
	$sameopp, # avoid previous opponent
	$pairsvr, # pair those due to start vs those due to reply
 	$thisrep, # minimize repeats
	$_[2], # to recover player ID
	)
      },
    # allowed opponent filter
    sub {
      ($exagony ? $_[0][$_[1]]->Team() ne $_[0][$_[2]]->Team() : 1) && 
      $_[0][$_[1]]->Repeats($_[0][$_[2]]->ID()) <= $repeats
      },
    # optional arguments to subs
    [],
    )) {
    $dp->Dirty(1);
    $dp->Update();
    $tournament->TellUser('idone');
    }
  else {
    $tournament->TellUser('epfail');
    }
  }

=back

=cut

=head1 BUGS

Makes inappropriate use of private functions in TSH::Player.

=cut

1;
